#!/bin/bash

BN=${0##*/}

cn=$(hostname --fqdn 2>/dev/null)
bits=2048
days=1
httpdconf=/tmp/cert.$$
dummy=dummy-
subdir=/conf

usage () {
    echo "Usage:"
    echo "  $BN [-h HOST] [-b BITS] [-d DAYS] [-c CONF_DIR] [-r]"
    echo
    echo "Options:"
    echo " -h host: certificate host name (default: $cn)"
    echo " -d days: certificate validation period in days (default: $days)"
    echo " -b bits: key size in bits (default: $bits)"
    echo " -c dir:  store certificates in apache conf directory (default: $httpdconf)"
    echo " -r:      create certificate named server.crt (not dummy-server.crt)"
    echo "Example:"
    echo "  $BN -d 365 -b 4092 -c /etc/httpd"
    exit 1
}

while getopts h:b:d:c:r c ; do
  case $c in
    h) cn=$OPTARG ;;
    b) bits=$OPTARG ;;
    d) days=$OPTARG ;;
    c) httpdconf=$OPTARG ;;
    r) dummy= ; subdir= ;;
    *) usage ;;
  esac
done

openssl version || exit 1

#-----------------------------------------------------------------------------
# define certificate subject
#-----------------------------------------------------------------------------

# cn == www.epages.de -> host == (de epages www)
host=( $(tr '.' '\n' <<< $cn | tac) )

declare -A cert
cert[countryName]=${host[0]}
cert[stateOrProvinceName]=${host[1]:-MyState}
cert[localityName]=${host[2]:-MyTown}
cert[organizationName]=${host[1]:-MyCompany}
cert[commonName]=$cn
cert[domainName]=${cn#*.}
[[ ${#cert[countryName]} = 2 ]] || cert[countryName]=US

algo=RSA
conf="$httpdconf$subdir/ssl"

mkdir -m 755 -p "$conf".crt "$conf".csr "$conf".key || exit 1

#-----------------------------------------------------------------------------
# exit if file(s) exist
#-----------------------------------------------------------------------------

for name in server ca ; do
  for file in crt csr key ; do
    ls -l "$conf.$file"/$name.$file 2>/dev/null && {
      echo "(existing file found, exiting.)"; exit 1
    }
  done
done

#-----------------------------------------------------------------------------
# find some random files
#-----------------------------------------------------------------------------

unset rand
for file in /etc/hosts /etc/group /etc/resolv.conf /bin/ls; do
    [[ -r $file ]] && rand="$rand:$file"
done

cfg=$httpdconf/.mkcert.cfg
rand="${rand#:}"
export RANDFILE=$httpdconf/.mkcert.rnd

#-----------------------------------------------------------------------
# ECHO
#-----------------------------------------------------------------------

ECHO () {
  local LINE_NUMBER=$(printf "%3d" $1)
  shift
  echo -e "[${0##*/}:$LINE_NUMBER] $@"
}

#-----------------------------------------------------------------------------
# create_req_cfg
#-----------------------------------------------------------------------------

create_req_cfg() {
  local unit=$1

  cat > $cfg <<!
[ req ]
default_bits            = $bits
distinguished_name      = req_DN
prompt                  = no
[ req_DN ]
countryName             = ${cert[countryName]}
stateOrProvinceName     = ${cert[stateOrProvinceName]}
localityName            = ${cert[localityName]}
organizationName        = ${cert[organizationName]}
organizationalUnitName  = $unit
commonName              = ${cert[commonName]}
emailAddress            = $unit@${cert[domainName]}
!
[[ $unit = unity ]] && cat >> $cfg <<!
subjectAltName          = DNS.1=*.$cn
!
}

#-----------------------------------------------------------------------------
# create_x509_cfg
#-----------------------------------------------------------------------------

create_x509_cfg() {
  local unit=$1

  cat > $cfg <<!
extensions = x509v3
[ x509v3 ]
subjectAltName   = email:copy
nsComment        = "mod_ssl generated test server certificate"
nsCertType       = $unit
!
[[ $unit = sslCA ]] && cat >> $cfg <<!
basicConstraints = CA:true,pathlen:0
!
}

#-----------------------------------------------------------------------------
# openssl_genrsa
#-----------------------------------------------------------------------------
openssl_genrsa() {
  local unit=$1
  local lineno=$2

  ECHO $lineno "openssl genrsa -rand $rand -out $conf.key/$unit.key $bits"
  openssl genrsa -rand $rand -out "$conf".key/$unit.key $bits || exit 1
}

#-----------------------------------------------------------------------------
#
#-----------------------------------------------------------------------------

openssl_req() {
  local unit=$1
  local lineno=$2

  ECHO $lineno "openssl req -config $cfg -new -key $conf.key/$unit.key -out $conf.csr/$unit.csr"
  openssl req -config $cfg -new -key "$conf".key/$unit.key -out "$conf".csr/$unit.csr || exit 1
}

#-----------------------------------------------------------------------------
#
#-----------------------------------------------------------------------------

openssl_x509() {
  local unit=$1
  local lineno=$2
  local serial=$3

  if [[ $unit = ${dummy}ca ]] ; then
    ECHO $lineno "openssl x509 -extfile $cfg -days $days -signkey $conf.key/$unit.key -in $conf.csr/$unit.csr -req -out $conf.crt/$unit.crt"
    openssl x509 -extfile $cfg -days $days -signkey "$conf".key/$unit.key \
      -in "$conf".csr/$unit.csr -req -out "$conf".crt/$unit.crt || exit 1
  else
    ECHO $lineno "openssl x509 -extfile $cfg -days $days -set_serial $serial -CA $conf.crt/${dummy}ca.crt -CAkey $conf.key/${dummy}ca.key -in $conf.csr/$unit.csr -req -out $conf.crt/$unit.crt"
    openssl x509 -extfile $cfg -days $days -set_serial $serial \
      -CA "$conf".crt/${dummy}ca.crt -CAkey "$conf".key/${dummy}ca.key \
      -in "$conf".csr/$unit.csr -req -out "$conf".crt/$unit.crt || exit 1
  fi

}

#-----------------------------------------------------------------------------
# create CA certificate
#-----------------------------------------------------------------------------

openssl_genrsa ${dummy}ca $LINENO
create_req_cfg ca
openssl_req ${dummy}ca $LINENO
create_x509_cfg sslCA
openssl_x509 ${dummy}ca $LINENO

#-----------------------------------------------------------------------------
# create server certificate
#-----------------------------------------------------------------------------
serial=$(date +%s)

openssl_genrsa ${dummy}server $LINENO
create_req_cfg server
openssl_req ${dummy}server $LINENO
create_x509_cfg server
openssl_x509 ${dummy}server $LINENO $serial

#-----------------------------------------------------------------------------
# create unity certificate
#-----------------------------------------------------------------------------
serial=$((serial + 1))

openssl_genrsa ${dummy}unity $LINENO
create_req_cfg unity
openssl_req ${dummy}unity $LINENO
create_x509_cfg server
openssl_x509 ${dummy}unity $LINENO $serial

#-----------------------------------------------------------------------------
# finish
#-----------------------------------------------------------------------------

chmod 600 $httpdconf$subdir/ssl.*/*
chmod 644 $httpdconf$subdir/ssl.crt/*

echo

find $httpdconf$subdir/ssl.* -type f -exec ls \-l {} \;
